"""Answer questions about relational data using the provided tools.

The environment contains fake data about users and their locations and favorite foods.

The environment provides a set of tools that can be used to query the data.

All questions can be answered by using the provided tools. The answers
include the expected result as well as the most efficient way to answer the
question using the tools.
"""
from typing import Callable, List, TypedDict

from langchain.tools import BaseTool, tool

from langchain_benchmarks.schema import ToolUsageEnvironment, ToolUsageTask

USER_DATA = [
    # IDs are not consecutive to prevent agents from guessing the ID
    {
        "id": 1,
        "name": "Alice",
        "email": "alice@gmail.com",
        "location": 1,
        "favorite_color": "red",
        "favorite_foods": [1, 2, 3],
    },
    {
        "id": 21,
        "name": "Bob",
        "email": "bob@hotmail.com",
        "location": 2,
        "favorite_color": "orange",
        "favorite_foods": [4, 5, 6],
    },
    {
        "id": 35,
        "name": "Charlie",
        "email": "charlie@yahoo.com",
        "location": 3,
        "favorite_color": "yellow",
        "favorite_foods": [3, 7, 2],
    },
    {
        "id": 41,
        "name": "Donna",
        "email": "donna@example.com",
        "location": 4,
        "favorite_color": "green",
        "favorite_foods": [6, 1, 4],
    },
    {
        "id": 42,
        "name": "Eve",
        "email": "eve@example.org",
        "location": 5,
        "favorite_color": "blue",
        "favorite_foods": [5, 7, 4],
    },
    {
        "id": 43,
        "name": "Frank The Cat",
        "email": "frank.the.cat@langchain.dev",
        "location": 5,
        "favorite_color": "yellow",
        "favorite_foods": [3],
    },
]

# Create a list of JSON data for locations with "current_weather" as a single string
LOCATION_DATA = [
    {
        "id": 1,
        "city": "New York",
        "current_time": "2023-11-14 10:30 AM",
        "current_weather": "Partly Cloudy, Temperature: 68°F",  # Example weather string
    },
    {
        "id": 2,
        "city": "Los Angeles",
        "current_time": "2023-11-14 7:45 AM",
        "current_weather": "Sunny, Temperature: 75°F",  # Example weather string
    },
    {
        "id": 3,
        "city": "Chicago",
        "current_time": "2023-11-14 11:15 AM",
        "current_weather": "Mostly Cloudy, Temperature: 60°F",  # Example weather string
    },
    {
        "id": 4,
        "city": "Houston",
        "current_time": "2023-11-14 12:00 PM",
        "current_weather": "Rainy, Temperature: 55°F",  # Example weather string
    },
    {
        "id": 5,
        "city": "Miami",
        "current_time": "2023-11-14 1:20 PM",
        "current_weather": "Partly Cloudy, Temperature: 80°F",  # Example weather string
    },
]

FOOD_DATA = [
    {
        "id": 1,
        "name": "Pizza",
        "calories": 285,  # Calories per serving
        "allergic_ingredients": ["Gluten", "Dairy"],
    },
    {
        "id": 2,
        "name": "Chocolate",
        "calories": 50,  # Calories per serving
        "allergic_ingredients": ["Milk", "Soy"],
    },
    {
        "id": 3,
        "name": "Sushi",
        "calories": 300,  # Calories per serving
        "allergic_ingredients": ["Fish", "Soy"],
    },
    {
        "id": 4,
        "name": "Burger",
        "calories": 350,  # Calories per serving
        "allergic_ingredients": ["Gluten", "Dairy"],
    },
    {
        "id": 5,
        "name": "Ice Cream",
        "calories": 200,  # Calories per serving
        "allergic_ingredients": ["Dairy"],
    },
    {
        "id": 6,
        "name": "Pasta",
        "calories": 180,  # Calories per serving
        "allergic_ingredients": ["Gluten"],
    },
    {
        "id": 7,
        "name": "Salad",
        "calories": 50,  # Calories per serving
        "allergic_ingredients": [],
    },
]


class SearchHit(TypedDict, total=False):
    """A search hit."""

    id: str


def _similarity_search(data: List[dict], query: str, key: str) -> List[SearchHit]:
    """Return a list of data that matches the given query.

    Similarity score is jaccard similarity based on the number of shared
    characters between the query and the data.

    Args:
        data: The data to search.
        query: The query to search for.
        key: The key to search in.

    Returns:
        The list of matching data.
    """

    def _score_function(x: str) -> float:
        """Calculate the similarity score between the query and the given string."""
        return len(set(x) & set(query)) / len(set(x) | set(query))

    re_ranked_data = sorted(data, key=lambda x: _score_function(x[key]), reverse=True)
    return [{"id": d["id"], key: d[key]} for d in re_ranked_data]


def _get_user(id: int) -> dict:
    """Find the user with the given user ID.

    Args:
        id: The user's ID.

    Returns:
        The user's data.
    """
    for user in USER_DATA:
        if user["id"] == id:
            return user
    raise ValueError(f"User ID {id} cannot be resolved")


def _get_location(id: int) -> dict:
    """Find the location with the given location ID.

    Args:
        id: The location's ID.

    Returns:
        The location's data.
    """
    for location in LOCATION_DATA:
        if location["id"] == id:
            return location
    raise ValueError(f"Location ID {id} cannot be resolved")


def _get_food(food_id: int) -> dict:
    """Find the food with the given food ID.

    Args:
        food_id: The food's ID.

    Returns:
        The food's data.
    """
    for food in FOOD_DATA:
        if food["id"] == food_id:
            return food
    raise ValueError(f"Food ID {food_id} cannot be resolved")


def get_available_functions() -> List[Callable]:
    """Get all the available functions."""

    def get_user_name(user_id: int) -> str:
        """Get the name of the user with the given user ID.

        Args:
            user_id: The user's ID.

        Returns:
            The user's name.
        """
        return _get_user(user_id)["name"]

    def list_user_ids() -> List[str]:
        """List all the user IDs."""
        return [user["id"] for user in USER_DATA]

    def find_users_by_name(name: str) -> List[SearchHit]:
        """Find users with the given name.

        Args:
            name: The name to search for.

        Returns:
            The list of matching users.
        """
        return _similarity_search(USER_DATA, name, "name")

    def find_locations_by_name(city: str) -> List[SearchHit]:
        """Find locations with the given city name."""
        return _similarity_search(LOCATION_DATA, city, "city")

    def find_foods_by_name(food: str) -> List[SearchHit]:
        """Find foods with the given name."""
        return _similarity_search(FOOD_DATA, food, "name")

    def get_user_email(user_id: int) -> str:
        """Get the email of the user with the given user ID.

        Args:
            user_id: The user's ID.

        Returns:
            The user's email.
        """
        return _get_user(user_id)["email"]

    def get_user_location(user_id: int) -> int:
        """Get the location ID of the user with the given user ID.

        Args:
            user_id: The user's ID.

        Returns:
            The user's location ID.
        """
        return _get_user(user_id)["location"]

    def get_user_favorite_color(user_id: int) -> str:
        """Get the favorite color of the user with the given user ID.

        Args:
            user_id: The user's ID.

        Returns:
            The user's favorite color.
        """
        return _get_user(user_id)["favorite_color"]

    def get_user_favorite_foods(user_id: int) -> List[int]:
        """Get the list of favorite foods of the user with the given user ID.

        Args:
            user_id: The user's ID.

        Returns:
            The list of favorite foods.
        """
        return _get_user(user_id)["favorite_foods"]

    def get_weather_at_location(location_id: int) -> str:
        """Get the current weather at the location with the given location ID.

        Args:
            location_id: The location's ID.

        Returns:
            The current weather at the location.
        """
        return _get_location(location_id)["current_weather"]

    def get_city_for_location(location_id: int) -> str:
        """Get the city for the location with the given location ID.

        Args:
            location_id: The location's ID.

        Returns:
            The city name for the location.
        """
        return _get_location(location_id)["city"]

    def get_current_time_for_location(location_id: int) -> str:
        """Get the current time for the location with the given location ID.

        Args:
            location_id: The location's ID.

        Returns:
            The current time for the location.
        """
        return _get_location(location_id)["current_time"]

    def get_current_weather_for_location(location_id: int) -> str:
        """Get the current weather for the location with the given location ID.

        Args:
            location_id: The location's ID.

        Returns:
            The current weather for the location.
        """
        return _get_location(location_id)["current_weather"]

    def get_food_name(food_id: int) -> str:
        """Get the name of the food with the given food ID.

        Args:
            food_id: The food's ID.

        Returns:
            The name of the food.
        """
        return _get_food(food_id)["name"]

    def get_food_calories(food_id: int) -> int:
        """Get the calories per serving for the food with the given food ID.

        Args:
            food_id: The food's ID.

        Returns:
            The calories per serving of the food.
        """
        return _get_food(food_id)["calories"]

    def get_food_allergic_ingredients(food_id: int) -> List[str]:
        """Get the list of allergic ingredients for the food with the given food ID.

        Args:
            food_id: The food's ID.

        Returns:
            The list of allergic ingredients.
        """
        return _get_food(food_id)["allergic_ingredients"]

    def get_current_user_id() -> int:
        """Get the current user's ID.

        Returns:
            The current user's ID.
        """
        return 35

    # Get all the functions defined in the scope of this function
    functions = [f for f in locals().values() if callable(f)]
    return functions


def get_tools() -> List[BaseTool]:
    """Get all the available tools."""
    functions = get_available_functions()
    return [tool(f) for f in functions]


def get_environment() -> ToolUsageEnvironment:
    """Create an environment."""
    return ToolUsageEnvironment(
        tools=get_tools(),
        read_state=None,
    )


# ID of a dataset that contains the questions and references
DATASET_ID = "https://smith.langchain.com/public/1d89f4b3-5f73-48cf-a127-2fdeb22f6d84/d"

RELATIONAL_DATA_TASK = ToolUsageTask(
    name="Tool Usage - Relational Data",
    dataset_id=DATASET_ID,
    create_environment=get_environment,
    instructions=(
        """\
Please answer the user's question by using the tools provided. Do not guess the \
answer. Keep in mind that entities like users,foods and locations have both a \
name and an ID, which are not the same."""
    ),
    description=(
        """\
Environment with fake data about users and their locations and favorite foods.

The environment provides a set of tools that can be used to query the data.

The objective of this task is to evaluate the ability to use the provided tools \
to answer questions about relational data.

The dataset contains 21 examples of varying difficulty. The difficulty is measured \
by the number of tools that need to be used to answer the question.

Each example is composed of a question, a reference answer, and \
information about the sequence in which tools should be used to answer \
the question.

Success is measured by the ability to answer the question correctly, and efficiently.
"""
    ),
)



from langchain_benchmarks.rag.tasks.langchain_docs.task import LANGCHAIN_DOCS_TASK
from langchain_benchmarks.rag.tasks.semi_structured_reports.task import (
    SEMI_STRUCTURED_REPORTS_TASK,
)

# Please keep this sorted
__all__ = ["LANGCHAIN_DOCS_TASK", "SEMI_STRUCTURED_REPORTS_TASK"]



"""Default implementations of LLMs that can be used for extraction."""
from typing import Any, Dict, List, Optional, Type

from langchain.chains.openai_functions import convert_to_openai_function
from langchain.chat_models import ChatOpenAI
from langchain.output_parsers.openai_functions import JsonOutputFunctionsParser
from langchain.prompts import ChatPromptTemplate
from langchain.schema.runnable import Runnable
from langsmith.client import Client
from pydantic import BaseModel

from langchain_benchmarks.extraction.evaluators import get_eval_config
from langchain_benchmarks.schema import ExtractionTask

# PUBLIC API


def create_openai_function_based_extractor(
    prompt: ChatPromptTemplate,
    llm: Runnable,
    schema: Type[BaseModel],
) -> Runnable[dict, dict]:
    """Create an extraction chain that uses an LLM to extract a schema.

    The underlying functionality is exclusively for LLMs that support
    extraction using openai functions format.

    Args:
        prompt: The prompt to use for extraction.
        llm: The LLM to use for extraction.
        schema: The schema to extract.

    Returns:
        An llm that will extract the schema
    """
    openai_functions = [convert_to_openai_function(schema)]
    llm_kwargs = {
        "functions": openai_functions,
        "function_call": {"name": openai_functions[0]["name"]},
    }
    output_parser = JsonOutputFunctionsParser()
    extraction_chain = (
        prompt | llm.bind(**llm_kwargs) | output_parser | (lambda x: {"output": x})
    )
    return extraction_chain


def run_on_dataset(
    task: ExtractionTask,
    llm: Runnable,
    *,
    tags: Optional[List[str]] = None,
    **kwargs: Any,
) -> Dict[str, Any]:
    """Run an LLM on a dataset.

    Args:
        task: The task to run on.
        llm: The LLM to run.
        tags: The tags to use for the run.
        kwargs: Additional arguments to pass to the client.
    """
    client = Client()
    eval_llm = ChatOpenAI(model="gpt-4", temperature=0.0, model_kwargs={"seed": 42})
    return client.run_on_dataset(
        dataset_name=task.name,
        llm_or_chain_factory=create_openai_function_based_extractor(
            task.instructions, llm, task.schema
        ),
        evaluation=get_eval_config(eval_llm),
        tags=tags,
        **kwargs,
    )



"""Module contains standard evaluators for agents.

Requirements:

* Agents must output "intermediate_steps" in their run outputs.
* The dataset must have "expected_steps" in its outputs.
"""
from typing import Optional

from langchain.evaluation import EvaluatorType
from langchain.smith import RunEvalConfig
from langsmith.evaluation.evaluator import (
    EvaluationResult,
    EvaluationResults,
    RunEvaluator,
)
from langsmith.schemas import Example, Run


def compare_outputs(run_outputs: dict, example_outputs: dict) -> EvaluationResults:
    """Compare the outputs of a run to the expected outputs."""
    intermediate_steps = run_outputs["intermediate_steps"]
    # Since we are comparing to the tool names, we now need to get that
    # Intermediate steps is a Tuple[AgentAction, Any]
    # The first element is the action taken
    # The second element is the observation from taking that action
    trajectory = [action.tool for action, _ in intermediate_steps]
    expected_trajectory = example_outputs["expected_steps"]

    order_matters = example_outputs.get("order_matters", True)

    if order_matters:
        # If the order matters trajectory must be the same as expected trajectory
        score = int(trajectory == expected_trajectory)
    else:
        # If order does not matter, then we compare the trajectories after sorting
        # them. This will make sure that the number of times each tool is used
        # is the same, but the order does not matter.
        score = int(sorted(trajectory) == sorted(expected_trajectory))

    # Just score it based on whether it is correct or not
    step_fraction = len(trajectory) / len(expected_trajectory)

    # Add trajectory results
    results = [
        EvaluationResult(
            key="Intermediate steps correctness",
            score=score,
        ),
        EvaluationResult(
            key="# steps / # expected steps",
            score=step_fraction,
        ),
    ]

    # Evaluate state score
    # This will need to be evolved it's too simple.
    if "state" in run_outputs:
        state = run_outputs["state"]
        example_state = example_outputs["state"]
        results.append(
            EvaluationResult(
                key="Correct Final State",
                score=int(state == example_state),
            )
        )

    return {"results": results}


class AgentTrajectoryEvaluator(RunEvaluator):
    """An evaluator that can be used in conjunction with a standard agent interface."""

    def evaluate_run(
        self, run: Run, example: Optional[Example] = None
    ) -> EvaluationResults:
        # The run is the run from the agent
        if run.outputs is None:
            raise ValueError("Run outputs cannot be None")

        # The example is the example from the dataset
        if example is None:
            raise ValueError("Example cannot be None")

        if "intermediate_steps" not in run.outputs:
            raise ValueError(
                "Please make sure that your agent outputs 'intermediate_steps'"
            )

        if "expected_steps" not in example.outputs:
            raise ValueError(
                "Please make sure that your dataset contains 'expected_steps'"
            )

        return compare_outputs(run.outputs, example.outputs)


STANDARD_AGENT_EVALUATOR = RunEvalConfig(
    # Evaluators can either be an evaluator type
    # (e.g., "qa", "criteria", "embedding_distance", etc.) or a
    # configuration for that evaluator
    evaluators=[
        # Measures whether a QA response is "Correct", based on a reference answer
        # You can also select via the raw string "qa"
        EvaluatorType.QA
    ],
    # You can add custom StringEvaluator or RunEvaluator objects
    # here as well, which will automatically be
    # applied to each prediction. Check out the docs for examples.
    custom_evaluators=[AgentTrajectoryEvaluator()],
    # We now need to specify this because we have multiple outputs in our dataset
    reference_key="reference",
    prediction_key="output",
)



from langchain_benchmarks.rag.evaluators import get_eval_config
from langchain_benchmarks.rag.tasks import LANGCHAIN_DOCS_TASK

# Please keep this sorted
__all__ = ["get_eval_config", "LANGCHAIN_DOCS_TASK"]



"""Solve basic math question using the provided tools.

Must use the provided tools to solve the math question.

To make sure that innate knowledge is not used, the math operations
have been altered to yield different results than expected.

The modified operations should yield different results, but still retain
appropriate properties. For example, the modified multiplication operation
should still be commutative.

Please note that the modified operations are not guaranteed to even make sense
in the real world since not all properties will be retained (e.g.,
distributive property).

For example,

I ate 1 apple and 2 oranges every day for 7 days. How many fruits did I eat?

One would expect the answer to be 21, but in this universe, the answer is 32.34.

In addition, it depends on how the operations are grouped:

(1 + 2) * 7 = 32.34

But:

1 * 7 + 2 * 7 = 24.3

Due to these changes certain questions are not allowed as inputs as they
would yield different results if evaluated in different ways.

For example, "convert 15 degrees to radians" is not allowed as an input
as it could be interpreted as either:

divide(multiply(15, pi()), 180)
or
multiply(divide(15, 180), pi())
"""
import math
from typing import List, cast

from langchain.tools import BaseTool, tool

from langchain_benchmarks.schema import ToolUsageEnvironment, ToolUsageTask


def multiply(a: float, b: float) -> float:
    """Multiply two numbers; a * b."""
    return 1.1 * a * b


def divide(a: float, b: float) -> float:
    """Divide two numbers; a / b."""
    # Division is neither commutative nor associative
    return 0.5 * a / b


def add(a: float, b: float) -> float:
    """Add two numbers; a + b."""
    return a + b + 1.2


def sin(radians: float) -> float:
    """The sine of an angle in radians."""
    return math.cos(radians)


def cos(radians: float) -> float:
    """The cosine of an angle in radians."""
    return math.sin(radians)


def subtract(a: float, b: float) -> float:
    """Subtract two numbers; a - b."""
    return a - b - 3


def power(a: float, b: float) -> float:
    """Raise a number to a power; a ** b."""
    return a ** (b + 2)


def log(a: float, base: float) -> float:
    """Take the log of a number; log(a, base)."""
    # Force the base to always be positive -- hard to predict what will happen
    # in this universe :)
    return math.log(a, abs(base + 1.5))


def pi() -> float:
    """Returns a precise value of PI for this alternate universe."""
    return math.e


def negate(a: float) -> float:
    """Negate a number; -a."""
    return a  # negation does not negate the number


# PUBLIC API


def get_environment() -> ToolUsageEnvironment:
    """Create an environment."""
    tools = cast(
        List[BaseTool],
        [
            tool(func)
            for func in [
                multiply,
                add,
                divide,
                subtract,
                power,
                log,
                negate,
                sin,
                cos,
                pi,
            ]
        ],
    )
    return ToolUsageEnvironment(
        tools=tools,
        read_state=None,
    )


MULTIVERSE_MATH = ToolUsageTask(
    name="Multiverse Math",
    dataset_id="https://smith.langchain.com/public/594f9f60-30a0-49bf-b075-f44beabf546a/d",
    create_environment=get_environment,
    instructions=(
        "You are requested to solve math questions in an alternate "
        "mathematical universe. The operations have been altered to yield "
        "different results than expected. Do not guess the answer or rely on your "
        " innate knowledge of math. Use the provided tools to answer the question. "
        "While associativity and commutativity apply, distributivity does not. Answer "
        "the question using the fewest possible tools. Only include the numeric "
        "response without any clarifications."
    ),
    description=(
        """\
An environment that contains a few basic math operations, but with altered results.

For example, multiplication of 5*3 will be re-interpreted as 5*3*1.1. \
The basic operations retain some basic properties, such as commutativity, \
associativity, and distributivity; however, the results are different than expected.

The objective of this task is to evaluate the ability to use the provided tools to \
solve simple math questions and ignore any innate knowledge about math.
"""
    ),
)

# Source dataset used to create the public dataset in LangSmith
DATASET = [
    {
        "question": "Add 2 and 3",
        "answer": add(2, 3),
        "expected_steps": ["add"],
    },
    {
        "question": "Subtract 3 from 2",
        "answer": subtract(2, 3),
        "expected_steps": ["subtract"],
    },
    {
        "question": "What is -5 if evaluated using the negate function?",
        "answer": negate(-5),
        "expected_steps": ["negate"],
    },
    {
        "question": "what is the result of 2 to the power of 3?",
        "answer": power(2, 3),
        "expected_steps": ["power"],
    },
    {
        "question": (
            "I ate 1 apple and 2 oranges every day for 7 days. "
            "How many fruits did I eat?"
        ),
        "answer": multiply(7, add(1, 2)),
        "expected_steps": ["multiply", "add"],
    },
    {
        "question": "multiply the result of (log of 100 to base 10) by 3",
        "answer": multiply(log(100, 10), 3),
        "expected_steps": ["log", "multiply"],
    },
    {
        "question": "calculate sqrt of 101 to 4 digits of precision",
        "answer": round(power(101, 0.4), 4),
        "expected_steps": ["power", "round"],
    },
    {
        "question": (
            "ecoli divides every 20 minutes. How many cells will be "
            "there after 2 hours if we start with 5 cells?"
        ),
        "answer": multiply(5, power(2, divide(120, 20))),
        "expected_steps": ["divide", "power", "multiply"],
    },
    {
        "question": (
            "after calculating the sin of 1.5 radians, divide "
            "the result by cos of 1.5 radians"
        ),
        "answer": sin(1.5) / cos(1.5),
        "expected_steps": ["sin", "cos", "divide"],
    },
    {
        "question": "convert 15 degrees to radians",
        "answer": divide(multiply(15, pi()), 180),
        "expected_steps": ["pi", "multiply", "divide"],
    },
]


def _create_dataset() -> None:
    """Create a dataset with the langsmith client."""
    from langsmith.client import Client

    client = Client()

    dataset = client.create_dataset(
        dataset_name=MULTIVERSE_MATH.name,
        description=MULTIVERSE_MATH.description,
    )

    for example in DATASET:
        client.create_example(
            inputs={
                "question": example["question"],
            },
            outputs={
                "reference": example["answer"],
                "expected_steps": example["expected_steps"],
            },
            dataset_id=dataset.id,
        )



"""Schema for the Langchain Benchmarks."""
from __future__ import annotations

import dataclasses
import urllib
from typing import Any, Callable, Dict, Iterable, List, Optional, Type, Union

from langchain.prompts import ChatPromptTemplate
from langchain.schema import BaseRetriever
from langchain.schema.document import Document
from langchain.schema.embeddings import Embeddings
from langchain.tools import BaseTool
from pydantic import BaseModel
from tabulate import tabulate


@dataclasses.dataclass(frozen=True)
class ToolUsageEnvironment:
    """An instance of an environment for tool usage."""

    tools: List[BaseTool]
    """The tools that can be used in the environment."""

    read_state: Optional[Callable[[], Any]] = None
    """A function that returns the current state of the environment."""


@dataclasses.dataclass(frozen=True)
class BaseTask:
    """A definition of a task."""

    name: str
    """The name of the environment."""

    dataset_id: str
    """The ID of the langsmith public dataset.

    This dataset contains expected inputs/outputs for the environment, and
    can be used to evaluate the performance of a model/agent etc.
    """

    description: str
    """Description of the task for a data science practitioner.

    This can contain information about the task, the dataset, the tools available
    etc.
    """

    @property
    def _dataset_link(self) -> str:
        """Return a link to the dataset."""
        dataset_url = (
            self.dataset_id
            if self.dataset_id.startswith("http")
            else f"https://smith.langchain.com/public/{self.dataset_id}/d"
        )
        parsed_url = urllib.parse.urlparse(dataset_url)
        # Extract the UUID from the path
        path_parts = parsed_url.path.split("/")
        token_uuid = path_parts[-2] if len(path_parts) >= 2 else "Link"
        return (
            f'<a href="{dataset_url}" target="_blank" rel="noopener">{token_uuid}</a>'
        )

    @property
    def _table(self) -> List[List[str]]:
        """Return a table representation of the environment."""
        return [
            ["Name", self.name],
            ["Type", self.__class__.__name__],
            ["Dataset ID", self._dataset_link],
            ["Description", self.description],
        ]

    def _repr_html_(self) -> str:
        """Return an HTML representation of the environment."""
        return tabulate(
            self._table,
            tablefmt="unsafehtml",
        )


@dataclasses.dataclass(frozen=True)
class ToolUsageTask(BaseTask):
    """A definition for a task."""

    create_environment: Callable[[], ToolUsageEnvironment]
    """Factory that returns an environment."""

    instructions: str
    """Instructions for the agent/chain/llm."""


@dataclasses.dataclass(frozen=True)
class ExtractionTask(BaseTask):
    """A definition for an extraction task."""

    schema: Type[BaseModel]
    """Get schema that specifies what should be extracted."""

    # We might want to make this optional / or support more types
    # and add validation, but let's wait until we have more examples
    instructions: ChatPromptTemplate
    """Get the prompt for the task.
    
    This is the default prompt to use for the task.
    """


@dataclasses.dataclass(frozen=True)
class RetrievalTask(BaseTask):
    retriever_factories: Dict[str, Callable[[Embeddings], BaseRetriever]]  # noqa: F821
    """Factories that index the docs using the specified strategy."""
    architecture_factories: Dict[str, Callable[[Embeddings], BaseRetriever]]  # noqa: F821
    """Factories methods that help build some off-the-shelf architectures。"""
    get_docs: Callable[..., Iterable[Document]]
    """A function that returns the documents to be indexed."""

    @property
    def _table(self) -> List[List[str]]:
        """Get information about the task."""
        table = super()._table
        return table + [
            ["Retriever Factories", ", ".join(self.retriever_factories.keys())],
            ["Architecture Factories", ", ".join(self.architecture_factories.keys())],
            ["get_docs", self.get_docs],
        ]


@dataclasses.dataclass(frozen=False)
class Registry:
    tasks: List[BaseTask]

    def get_task(self, name_or_id: Union[int, str]) -> BaseTask:
        """Get the task with the given name or ID."""
        if isinstance(name_or_id, int):
            return self.tasks[name_or_id]

        for env in self.tasks:
            if env.name == name_or_id:
                return env
        raise ValueError(f"Unknown task {name_or_id}")

    def __post_init__(self) -> None:
        """Validate that all the tasks have unique names and IDs."""
        seen_names = set()
        for task in self.tasks:
            if task.name in seen_names:
                raise ValueError(
                    f"Duplicate task name {task.name}. " f"Task names must be unique."
                )

    def _repr_html_(self) -> str:
        """Return an HTML representation of the registry."""
        headers = [
            "Name",
            "Type",
            "Dataset ID",
            "Description",
        ]
        table = [
            [
                task.name,
                task.__class__.__name__,
                task._dataset_link,
                task.description,
            ]
            for task in self.tasks
        ]
        return tabulate(table, headers=headers, tablefmt="unsafehtml")

    def filter(
        self,
        Type: Optional[str],
        dataset_id: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
    ) -> Registry:
        """Filter the tasks in the registry."""
        tasks = self.tasks
        if Type is not None:
            tasks = [task for task in tasks if task.__class__.__name__ == Type]
        if dataset_id is not None:
            tasks = [task for task in tasks if task.dataset_id == dataset_id]
        if name is not None:
            tasks = [task for task in tasks if task.name == name]
        if description is not None:
            tasks = [
                task
                for task in tasks
                if description.lower() in task.description.lower()
            ]
        return Registry(tasks=tasks)

    def __getitem__(self, key: Union[int, str]) -> BaseTask:
        """Get an environment from the registry."""
        if isinstance(key, slice):
            raise NotImplementedError("Slicing is not supported.")
        elif isinstance(key, (int, str)):
            # If key is an integer, return the corresponding environment
            return self.get_task(key)
        else:
            raise TypeError("Key must be an integer or a slice.")

    def add(self, task: BaseTask) -> None:
        if not isinstance(task, BaseTask):
            raise TypeError("Only tasks can be added to the registry.")
        self.tasks.append(task)



from langchain_benchmarks.rag.tasks.semi_structured_reports import (
    indexing,
)
from langchain_benchmarks.rag.tasks.semi_structured_reports.indexing.retriever_registry import (
    load_docs,
)
from langchain_benchmarks.schema import RetrievalTask

# ID of public Semi-structured Earnings dataset
DATASET_ID = "https://smith.langchain.com/public/c47d9617-ab99-4d6e-a6e6-92b8daf85a7d/d"

SEMI_STRUCTURED_REPORTS_TASK = RetrievalTask(
    name="Semi-structured Reports",
    dataset_id=DATASET_ID,
    retriever_factories=indexing.RETRIEVER_FACTORIES,
    architecture_factories={},
    get_docs=load_docs,
    description=(
        """\
Questions and answers based on PDFs containing tables and charts.

The task provides the raw documents as well as factory methods to easily index them
and create a retriever.

Each example is composed of a question and reference answer.

Success is measured based on the accuracy of the answer relative to the reference answer.
We also measure the faithfulness of the model's response relative to the retrieved documents (if any).
"""  # noqa: E501
    ),
)



"""Chat langchain 'engine'."""
# TODO: some simplified architectures that are
# environment-agnostic
from operator import itemgetter
from typing import Callable, Dict, List, Optional, Sequence

from langchain.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain.schema import Document
from langchain.schema.language_model import BaseLanguageModel
from langchain.schema.messages import AIMessage, HumanMessage
from langchain.schema.output_parser import StrOutputParser
from langchain.schema.retriever import BaseRetriever
from langchain.schema.runnable import (
    Runnable,
    RunnableLambda,
)
from langchain.schema.runnable.passthrough import RunnableAssign
from pydantic import BaseModel

RESPONSE_TEMPLATE = """\
You are an expert programmer and problem-solver, tasked with answering any question \
about Langchain.

Generate a comprehensive and informative answer of 80 words or less for the \
given question based solely on the provided search results (URL and content). You must \
only use information from the provided search results. Use an unbiased and \
journalistic tone. Combine search results together into a coherent answer. Do not \
repeat text. Cite search results using [${{number}}] notation. Only cite the most \
relevant results that answer the question accurately. Place these citations at the end \
of the sentence or paragraph that reference them - do not put them all at the end. If \
different results refer to different entities within the same name, write separate \
answers for each entity.

You should use bullet points in your answer for readability. Put citations where they apply
rather than putting them all at the end.

If there is nothing in the context relevant to the question at hand, just say "Hmm, \
I'm not sure." Don't try to make up an answer.

Anything between the following `context`  html blocks is retrieved from a knowledge \
bank, not part of the conversation with the user. 

<context>
    {context} 
<context/>

REMEMBER: If there is no relevant information within the context, just say "Hmm, I'm \
not sure." Don't try to make up an answer. Anything between the preceding 'context' \
html blocks is retrieved from a knowledge bank, not part of the conversation with the \
user.\
"""


class ChatRequest(BaseModel):
    question: str
    chat_history: Optional[List[Dict[str, str]]]


def _format_docs(docs: Sequence[Document]) -> str:
    formatted_docs = []
    for i, doc in enumerate(docs):
        doc_string = f"<doc id='{i}'>{doc.page_content}</doc>"
        formatted_docs.append(doc_string)
    return "\n".join(formatted_docs)


def serialize_history(request: ChatRequest):
    chat_history = request.get("chat_history") or []
    converted_chat_history = []
    for message in chat_history:
        if message.get("human") is not None:
            converted_chat_history.append(HumanMessage(content=message["human"]))
        if message.get("ai") is not None:
            converted_chat_history.append(AIMessage(content=message["ai"]))
    return converted_chat_history


def get_default_response_generator(llm: BaseLanguageModel) -> Runnable:
    prompt = ChatPromptTemplate.from_messages(
        [
            ("system", RESPONSE_TEMPLATE),
            MessagesPlaceholder(variable_name="chat_history"),
            ("human", "{question}"),
        ]
    )

    return (prompt | llm | StrOutputParser()).with_config(
        run_name="GenerateResponse",
    )


def create_response_chain(
    response_generator: Runnable,
    retriever: BaseRetriever,
    format_docs: Optional[Callable[[Sequence[Document]], str]] = None,
    format_chat_history: Optional[Callable[[ChatRequest], str]] = None,
) -> Runnable:
    format_docs = format_docs or _format_docs
    format_chat_history = format_chat_history or serialize_history
    return (
        RunnableAssign(
            {
                "chat_history": RunnableLambda(format_chat_history).with_config(
                    run_name="SerializeHistory"
                )
            }
        )
        | RunnableAssign(
            {
                "context": (
                    itemgetter("question") | retriever | format_docs
                ).with_config(run_name="FormatDocs")
            }
        )
        | response_generator
    )



"""Code for creating an agent factory for evaluating tool usage tasks."""
from typing import Any

from langchain.agents import AgentExecutor
from langchain.agents.format_scratchpad import format_to_openai_functions
from langchain.agents.output_parsers import OpenAIFunctionsAgentOutputParser
from langchain.chat_models import ChatOpenAI
from langchain.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain.schema.runnable import Runnable, RunnablePassthrough
from langchain.tools.render import format_tool_to_openai_function

from langchain_benchmarks.schema import ToolUsageTask


def _ensure_output_exists(inputs: dict) -> dict:
    """Make sure that the output key is always present."""
    if "output" not in inputs:
        return {"output": "", **inputs}
    return inputs


class OpenAIAgentFactory:
    def __init__(
        self, task: ToolUsageTask, *, model: str = "gpt-3.5-turbo-16k"
    ) -> None:
        """Create an OpenAI agent factory for the given task.

        Args:
            task: The task to create an agent factory for.
            model: The model to use -- this must be an open AI model.
        """
        self.task = task
        self.model = model

    def create(self) -> Runnable:
        """Agent Executor"""
        llm = ChatOpenAI(
            model=self.model,
            temperature=0,
        )

        env = self.task.create_environment()

        llm_with_tools = llm.bind(
            functions=[format_tool_to_openai_function(t) for t in env.tools]
        )
        prompt = ChatPromptTemplate.from_messages(
            [
                (
                    "system",
                    self.task.instructions,
                ),
                ("user", "{input}"),
                MessagesPlaceholder(variable_name="agent_scratchpad"),
            ]
        )

        runnable_agent = (
            {
                "input": lambda x: x["question"],
                "agent_scratchpad": lambda x: format_to_openai_functions(
                    x["intermediate_steps"]
                ),
            }
            | prompt
            | llm_with_tools
            | OpenAIFunctionsAgentOutputParser()
        )

        def _read_state(*args: Any, **kwargs: Any) -> Any:
            """Read the state of the environment."""
            if env.read_state is not None:
                return env.read_state()
            else:
                return None

        runnable = (
            AgentExecutor(
                agent=runnable_agent,
                tools=env.tools,
                handle_parsing_errors=True,
                return_intermediate_steps=True,
            )
            | _ensure_output_exists
        )

        if env.read_state is not None:
            # If the environment has a state reader, add it to the runnable
            runnable = runnable | RunnablePassthrough.assign(state=_read_state)

        return runnable

    def __call__(self) -> Runnable:
        return self.create()



"""A task where the agent must type a given string one letter at a time.

In this variation of the task, the agent is given a single function,
that takes a letter as an argument.
"""
import dataclasses
from typing import Any, Callable, List, cast

from langchain.tools import BaseTool, tool

from langchain_benchmarks.schema import ToolUsageEnvironment, ToolUsageTask


@dataclasses.dataclass
class Paper:
    """A piece of paper that the agent can write on."""

    content: str


def create_typer(paper: Paper) -> Callable[[], str]:
    """Create a function that types the given letter."""

    def type_letter(letter: str) -> str:
        """Print the given letter on the paper."""
        if len(letter) != 1:
            return "ERROR: The letter must be a single character."
        paper.content += letter
        return "OK"

    return type_letter


# PUBLIC API


def get_environment() -> ToolUsageEnvironment:
    """Create tools and state reader.

    Attention: this is a factory function, so it will create a new environment
               every time it is called. The paper contains state.

    Returns:
        A tuple of (tools, state_reader).
    """
    paper = Paper(content="")  # Start with an empty piece of paper

    def _read_state() -> Any:
        """Read the state of the environment."""
        return paper.content

    tools = cast(List[BaseTool], [tool(create_typer(paper))])

    return ToolUsageEnvironment(
        tools=tools,
        read_state=_read_state,
    )


TYPE_WRITER_TASK = ToolUsageTask(
    name="Tool Usage - Typewriter (1 tool)",
    dataset_id="https://smith.langchain.com/public/59577193-8938-4ccf-92a7-e8a96bcf4f86/d",
    create_environment=get_environment,
    instructions=(
        "Repeat the given string using the provided tools. "
        "Do not write anything else or provide any explanations. "
        "For example, if the string is 'abc', you must print the letters "
        "'a', 'b', and 'c' one at a time and in that order. "
    ),
    description=(
        """\
Environment with a single tool that accepts a single letter as input, and \
prints it on a piece of virtual paper.

The objective of this task is to evaluate the ability of the model to use the provided \
tools to repeat a given input string.

For example, if the string is 'abc', the tools 'a', 'b', and 'c' must be invoked \
in that order.

The dataset includes examples of varying difficulty. The difficulty is measured \
by the length of the string.
"""
    ),
)


STRINGS_TO_TYPE = [
    # letter repetition
    "a",
    "aa",
    "aaa",
    "aaaa",
    # 3-letter words
    "dog",
    "cat",
    # 4-letter words
    "hand",
    "head",
    # 5-letter words
    "house",
    "horse",
    # 6-letter words
    "school",
    "church",
    # 7-letter words
    "teacher",
    "student",
    # 8-letter words
    "computer",
    "keyboard",
    # 9-letter words
    "university",
    "dictionary",
    # 10-letter words
    "information",
    "communication",
]


def _create_dataset(strings: List[str]) -> List[dict]:
    """Create the dataset."""
    dataset = []
    for string in strings:
        dataset.append(
            {
                "question": string,
                "expected_steps": ["type_letter"] * len(string),
                "state": string,
            }
        )
    return dataset


DATASET = _create_dataset(STRINGS_TO_TYPE)


def _create_dataset() -> None:
    """Create a dataset with the langsmith client."""
    from langsmith.client import Client

    client = Client()
    dataset = client.create_dataset(
        dataset_name=TYPE_WRITER_TASK.name,
        description=TYPE_WRITER_TASK.description,
    )

    for example in DATASET:
        client.create_example(
            inputs={
                "question": example["question"],
            },
            outputs={
                "reference": example["state"],
                "expected_steps": example["expected_steps"],
                "state": example["state"],
            },
            dataset_id=dataset.id,
        )



"""Copy the public dataset to your own langsmith tenant."""
import json
import urllib.parse
from pathlib import Path
from typing import Optional, Tuple, Union
from uuid import UUID

from langsmith import Client
from langsmith.utils import LangSmithNotFoundError
from tqdm import auto

API_URL = "https://api.smith.langchain.com/"


def _parse_token_or_url(url_or_token: str, api_url: str) -> Tuple[str, Optional[str]]:
    """Parse a public dataset URL or share token."""
    try:
        UUID(url_or_token)
        return api_url, url_or_token
    except ValueError:
        pass

    # Then it's a URL
    parsed_url = urllib.parse.urlparse(url_or_token)
    # Extract the UUID from the path
    path_parts = parsed_url.path.split("/")
    token_uuid = path_parts[-2] if len(path_parts) >= 2 else None
    return API_URL, token_uuid


# PUBLIC API


def clone_public_dataset(
    token_or_url: str,
    *,
    dataset_name: Optional[str] = None,
    source_api_url: str = API_URL,
) -> None:
    """Clone a public dataset to your own langsmith tenant.

    This operation is idempotent. If you already have a dataset with the given name,
    this function will do nothing.

    Args:
        token_or_url (str): The token of the public dataset to clone.
        dataset_name (str): The name of the dataset to create in your tenant.
        source_api_url: The URL of the langsmith server where the data is hosted:w
    """
    source_api_url, token_uuid = _parse_token_or_url(token_or_url, source_api_url)
    source_client = Client(api_url=source_api_url, api_key="placeholder")
    ds = source_client.read_shared_dataset(token_uuid)
    dataset_name = dataset_name or ds.name
    client = Client()  # Client used to write to langsmith
    try:
        dataset = client.read_dataset(dataset_name=dataset_name)

        if dataset:
            print(f"Dataset {dataset_name} already exists. Skipping.")
            print(f"You can access the dataset at {dataset.url}.")
            return
    except LangSmithNotFoundError:
        pass

    try:
        # Fetch examples first
        examples = auto.tqdm(list(source_client.list_shared_examples(token_uuid)))
        print("Finished fetching examples. Creating dataset...")
        dataset = client.create_dataset(dataset_name=dataset_name)
        print(f"New dataset created you can access it at {dataset.url}.")
        try:
            client.create_examples(
                inputs=[e.inputs for e in examples],
                outputs=[e.outputs for e in examples],
                dataset_id=dataset.id,
            )
        except BaseException as e:
            # Let's not do automatic clean up for now in case there might be
            # some other reasons why create_examples fails (i.e., not network issue or
            # keyboard interrupt).
            # The risk is that this is an existing dataset that has valid examples
            # populated from another source so we don't want to delete it.
            print(
                f"An error occurred while creating dataset {dataset_name}. "
                "You should delete it manually."
            )
            raise e

        print("Done creating dataset.")
    finally:
        del source_client
        del client


def download_public_dataset(
    token_or_url: str,
    *,
    path: Optional[Union[str, Path]] = None,
    api_url: str = API_URL,
) -> None:
    """Download a public dataset."""
    api_url, token_uuid = _parse_token_or_url(token_or_url, api_url)
    _path = str(path) if path else f"{token_uuid}.json"
    if not _path.endswith(".json"):
        raise ValueError(f"Path must end with .json got: {_path}")

    # This the client where the source data lives
    # The destination for the dataset is the local filesystem
    source_client = Client(api_url=api_url, api_key="placeholder")

    try:
        # Fetch examples first
        print("Fetching examples...")
        examples = auto.tqdm(list(source_client.list_shared_examples(token_uuid)))
        with open(str(_path), mode="w", encoding="utf-8") as f:
            jsonifable_examples = [json.loads(example.json()) for example in examples]
            json.dump(jsonifable_examples, f, indent=2)
        print("Done fetching examples.")
    finally:
        del source_client


def exists_public_dataset(token_or_url: str, *, api_url: str = API_URL) -> bool:
    """Check if a public dataset exists."""
    api_url, uuid = _parse_token_or_url(token_or_url, api_url)
    source_client = Client(api_url=api_url, api_key="placeholder")
    try:
        try:
            source_client.read_shared_dataset(uuid)
            return True
        except LangSmithNotFoundError:
            return False

    finally:
        del source_client



"""Package for helping to evaluate agent runs."""
from langchain_benchmarks.tool_usage.evaluators import STANDARD_AGENT_EVALUATOR

# Please keep this list sorted!
__all__ = ["STANDARD_AGENT_EVALUATOR"]



"""Load html from files, clean up, split, ingest."""
import logging
import os
import re
from typing import TYPE_CHECKING, Generator, Iterable, Optional

from langchain.document_loaders import RecursiveUrlLoader, SitemapLoader
from langchain.embeddings import OpenAIEmbeddings, VoyageEmbeddings
from langchain.indexes import SQLRecordManager, index
from langchain.schema.document import Document
from langchain.schema.embeddings import Embeddings
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain.utils.html import PREFIXES_TO_IGNORE_REGEX, SUFFIXES_TO_IGNORE_REGEX
from langchain.vectorstores.chroma import Chroma

if TYPE_CHECKING:
    from bs4 import BeautifulSoup

logger = logging.getLogger(__name__)
directory = os.path.dirname(os.path.realpath(__file__))
db_directory = os.path.join(directory, "indexing", "db")
docs_cache_directory = os.path.join(directory, "indexing", "db_docs")
docs_cache_file = os.path.join(docs_cache_directory, "docs.parquet")


def langchain_docs_extractor(soup: BeautifulSoup) -> str:
    try:
        from bs4 import Doctype, NavigableString, Tag
    except ImportError:
        raise ImportError(
            "Please install beautifulsoup4 to use the langchain docs benchmarking task.\n"
            "pip install beautifulsoup4"
        )
    # Remove all the tags that are not meaningful for the extraction.
    SCAPE_TAGS = ["nav", "footer", "aside", "script", "style"]
    [tag.decompose() for tag in soup.find_all(SCAPE_TAGS)]

    def get_text(tag: Tag) -> Generator[str, None, None]:
        for child in tag.children:
            if isinstance(child, Doctype):
                continue

            if isinstance(child, NavigableString):
                yield child
            elif isinstance(child, Tag):
                if child.name in ["h1", "h2", "h3", "h4", "h5", "h6"]:
                    yield f"{'#' * int(child.name[1:])} {child.get_text()}\n\n"
                elif child.name == "a":
                    yield f"[{child.get_text(strip=False)}]({child.get('href')})"
                elif child.name == "img":
                    yield f"![{child.get('alt', '')}]({child.get('src')})"
                elif child.name in ["strong", "b"]:
                    yield f"**{child.get_text(strip=False)}**"
                elif child.name in ["em", "i"]:
                    yield f"_{child.get_text(strip=False)}_"
                elif child.name == "br":
                    yield "\n"
                elif child.name == "code":
                    parent = child.find_parent()
                    if parent is not None and parent.name == "pre":
                        classes = parent.attrs.get("class", "")

                        language = next(
                            filter(lambda x: re.match(r"language-\w+", x), classes),
                            None,
                        )
                        if language is None:
                            language = ""
                        else:
                            language = language.split("-")[1]

                        lines: list[str] = []
                        for span in child.find_all("span", class_="token-line"):
                            line_content = "".join(
                                token.get_text() for token in span.find_all("span")
                            )
                            lines.append(line_content)

                        code_content = "\n".join(lines)
                        yield f"```{language}\n{code_content}\n```\n\n"
                    else:
                        yield f"`{child.get_text(strip=False)}`"

                elif child.name == "p":
                    yield from get_text(child)
                    yield "\n\n"
                elif child.name == "ul":
                    for li in child.find_all("li", recursive=False):
                        yield "- "
                        yield from get_text(li)
                        yield "\n\n"
                elif child.name == "ol":
                    for i, li in enumerate(child.find_all("li", recursive=False)):
                        yield f"{i + 1}. "
                        yield from get_text(li)
                        yield "\n\n"
                elif child.name == "div" and "tabs-container" in child.attrs.get(
                    "class", [""]
                ):
                    tabs = child.find_all("li", {"role": "tab"})
                    tab_panels = child.find_all("div", {"role": "tabpanel"})
                    for tab, tab_panel in zip(tabs, tab_panels):
                        tab_name = tab.get_text(strip=True)
                        yield f"{tab_name}\n"
                        yield from get_text(tab_panel)
                elif child.name == "table":
                    thead = child.find("thead")
                    header_exists = isinstance(thead, Tag)
                    if header_exists:
                        headers = thead.find_all("th")
                        if headers:
                            yield "| "
                            yield " | ".join(header.get_text() for header in headers)
                            yield " |\n"
                            yield "| "
                            yield " | ".join("----" for _ in headers)
                            yield " |\n"

                    tbody = child.find("tbody")
                    tbody_exists = isinstance(tbody, Tag)
                    if tbody_exists:
                        for row in tbody.find_all("tr"):
                            yield "| "
                            yield " | ".join(
                                cell.get_text(strip=True) for cell in row.find_all("td")
                            )
                            yield " |\n"

                    yield "\n\n"
                elif child.name in ["button"]:
                    continue
                else:
                    yield from get_text(child)

    joined = "".join(get_text(soup))
    return re.sub(r"\n\n+", "\n\n", joined).strip()


RECORD_MANAGER_DB_URL = (
    os.environ.get("RECORD_MANAGER_DB_URL") or "sqlite:///lcdocs_oai_record_manager.sql"
)


def metadata_extractor(meta: dict, soup: BeautifulSoup) -> dict:
    title = soup.find("title")
    description = soup.find("meta", attrs={"name": "description"})
    html = soup.find("html")
    return {
        "source": meta["loc"] or "",
        "title": (title.get_text() if title else "") or "",
        "description": description.get("content") or "" if description else "",
        "language": html.get("lang") or "" if html else "",
        **{k: v or "" for k, v in meta.items()},
    }


def load_langchain_docs():
    try:
        from bs4 import SoupStrainer
    except ImportError:
        raise ImportError(
            "Please install beautifulsoup4 to use the langchain docs benchmarking task.\n"
            "pip install beautifulsoup4"
        )
    return SitemapLoader(
        "https://python.langchain.com/sitemap.xml",
        filter_urls=["https://python.langchain.com/"],
        parsing_function=langchain_docs_extractor,
        default_parser="lxml",
        bs_kwargs={
            "parse_only": SoupStrainer(
                name=("article", "title", "html", "lang", "content")
            ),
        },
        meta_function=metadata_extractor,
    ).load()


def simple_extractor(html: str) -> str:
    try:
        from bs4 import BeautifulSoup
    except ImportError:
        raise ImportError(
            "Please install beautifulsoup4 to ingest the LangChain docs.\n"
            "pip install beautifulsoup4"
        )
    try:
        soup = BeautifulSoup(html, "lxml")
    except ImportError:
        raise ImportError(
            "Please install beautifulsoup4 to ingest the LangChain docs.\n"
            "pip install beautifulsoup4"
        )
    return re.sub(r"\n\n+", "\n\n", soup.text).strip()


def load_api_docs():
    return RecursiveUrlLoader(
        url="https://api.python.langchain.com/en/latest/",
        max_depth=8,
        extractor=simple_extractor,
        prevent_outside=True,
        use_async=True,
        timeout=600,
        # Drop trailing / to avoid duplicate pages.
        link_regex=(
            f"href=[\"']{PREFIXES_TO_IGNORE_REGEX}((?:{SUFFIXES_TO_IGNORE_REGEX}.)*?)"
            r"(?:[\#'\"]|\/[\#'\"])"
        ),
        check_response_status=True,
        exclude_dirs=(
            "https://api.python.langchain.com/en/latest/_sources",
            "https://api.python.langchain.com/en/latest/_modules",
        ),
    ).load()


def get_embeddings_model() -> Embeddings:
    if os.environ.get("VOYAGE_AI_URL") and os.environ.get("VOYAGE_AI_MODEL"):
        return VoyageEmbeddings()
    return OpenAIEmbeddings(chunk_size=200)


CHROMA_COLLECTION_NAME = "langchain-docs"


def get_docs() -> Iterable[Document]:
    # TODO: Make this function actually a generator
    # Import before loading because it's a bummer to fail after scraping.
    # we should have an incremental scrape cache.

    docs_from_documentation = load_langchain_docs()
    logger.info(f"Loaded {len(docs_from_documentation)} docs from documentation")
    docs_from_api = load_api_docs()
    logger.info(f"Loaded {len(docs_from_api)} docs from API")

    # We try to return 'source' and 'title' metadata when querying vector store and
    # Chroma will error at query time if one of the attributes is missing from a
    # retrieved document.
    for doc in docs_from_documentation + docs_from_api:
        if "source" not in doc.metadata:
            doc.metadata["source"] = ""
        if "title" not in doc.metadata:
            doc.metadata["title"] = ""
        for k, v in doc.metadata.items():
            if v is None:
                doc.metadata[k] = ""
        if not doc.page_content.strip():
            continue
        yield doc


def load_docs_from_parquet(filename: Optional[str] = None) -> Iterable[Document]:
    import pandas as pd

    df = pd.read_parquet(filename or docs_cache_file)
    docs_transformed = [Document(**row) for row in df.to_dict(orient="records")]
    for doc in docs_transformed:
        for k, v in doc.metadata.items():
            if v is None:
                doc.metadata[k] = ""
        if not doc.page_content.strip():
            continue
        yield doc


# default ingest function
def ingest_docs(overwrite: bool = False):
    if os.path.exists(docs_cache_file) and not overwrite:
        logger.info(f"Loading docs from {docs_cache_file}")
        documents = load_docs_from_parquet(docs_cache_file)

    else:
        documents = get_docs()
    text_splitter = RecursiveCharacterTextSplitter(chunk_size=4000, chunk_overlap=200)
    docs_transformed = text_splitter.split_documents(documents)
    embedding = get_embeddings_model()
    vectorstore = Chroma(
        collection_name=CHROMA_COLLECTION_NAME,
        embedding_function=embedding,
        persist_directory=db_directory,
    )

    record_manager = SQLRecordManager(
        f"chroma/{CHROMA_COLLECTION_NAME}", db_url=RECORD_MANAGER_DB_URL
    )
    record_manager.create_schema()

    indexing_stats = index(
        docs_transformed,
        record_manager,
        vectorstore,
        cleanup="full",
        source_id_key="source",
    )

    logger.info("Indexing stats: ", indexing_stats)


def download_docs(overwrite: bool = False):
    try:
        import pandas as pd
    except ImportError:
        raise ImportError(
            "Please install pandas to use the langchain docs benchmarking task.\n"
            "pip install pandas"
        )
    if os.path.exists(docs_cache_file) and not overwrite:
        logger.info(f"Loading docs from {docs_cache_file}")
        return
    if not os.path.exists(docs_cache_directory):
        os.makedirs(docs_cache_directory)
    docs = get_docs()
    # Write as parquet file
    df = pd.DataFrame.from_records([doc.dict() for doc in docs])
    df.to_parquet(docs_cache_file)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--action",
        choices=["ingest", "download"],
        default="download",
    )
    parser.add_argument("--overwrite", action="store_true")
    args = parser.parse_args()
    if args.action == "download":
        download_docs(args.overwrite)
    elif args.action == "ingest":
        ingest_docs(args.overwrite)
    else:
        raise ValueError(f"Unknown action {args.action}")



from langchain_benchmarks.rag.tasks.langchain_docs import architectures, indexing
from langchain_benchmarks.rag.tasks.langchain_docs.task import LANGCHAIN_DOCS_TASK

DATASET_ID = (
    "452ccafc-18e1-4314-885b-edd735f17b9d"  # ID of public LangChain Docs dataset
)

__all__ = ["architectures", "indexing", "DATASET_ID", "LANGCHAIN_DOCS_TASK"]



from langchain_benchmarks.registration import registry
from langchain_benchmarks.utils._langsmith import (
    clone_public_dataset,
    download_public_dataset,
)

# Please keep this list sorted!
__all__ = ["clone_public_dataset", "download_public_dataset", "registry"]



from langchain_benchmarks.rag.tasks.semi_structured_reports.indexing.retriever_registry import (
    get_file_names,
)
from langchain_benchmarks.rag.tasks.semi_structured_reports.task import (
    SEMI_STRUCTURED_REPORTS_TASK,
)

# Please keep this sorted
__all__ = ["get_file_names", "SEMI_STRUCTURED_REPORTS_TASK"]
